"""Skill data services."""

from pathlib import Path
from typing import Iterable
from uuid import UUID

from ..models.skill import MatchSkill, PlaylistCsr
from ..xuid import wrap_xuid
from .base import BaseService

_HOST = "https://skill.svc.halowaypoint.com:443"


class SkillService(BaseService):
    """Skill data services."""

    async def get_match_skill(
        self, match_id: str | UUID, xuids: Iterable[str | int]
    ) -> MatchSkill:
        """Get player CSR and team MMR values for a given match and player list.

        Args:
            match_id: Halo Infinite match ID.
            xuids: The Xbox Live IDs of the match's players. Only
                players in this list will have their skill data returned.

        Returns:
            The skill data for the match.

        Raises:
            TypeError: If `xuids` is a `str` instead of an iterable of XUIDs.
            ValueError: If `xuids` is empty.
        """
        if isinstance(xuids, str):
            raise TypeError("`xuids` must be an iterable of XUIDs, got `str`")
        if not xuids:
            raise ValueError("`xuids` cannot be empty")
        url = f"{_HOST}/hi/matches/{match_id}/skill"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        return MatchSkill(**await self._get_json(url, params=params))

    async def get_playlist_csr(
        self,
        playlist_id: str | UUID,
        xuids: Iterable[str | int],
        season_id: str | None = None,
    ) -> PlaylistCsr:
        """Get player CSR values for a given playlist and player list.

        Args:
            playlist_id: Halo Infinite playlist asset ID.
            xuids: The Xbox Live IDs of the players.
            season_id: Halo Infinite season ID. If not provided, the
                current season will be used. Value can be provided as a path,
                e.g. "Csr/Seasons/CsrSeason5-1.json", or as a bare ID,
                e.g. "CsrSeason5-1". If provided as a path, the path prefix
                and ".json" extension will be removed. Case is not important.

        Returns:
            The summary CSR data for the players in the given playlist.

        Raises:
            TypeError: If `xuids` is a `str` instead of an iterable of XUIDs.
            ValueError: If `xuids` is empty.
        """
        if isinstance(xuids, str):
            raise TypeError("`xuids` must be an iterable of XUIDs, got `str`")
        if not xuids:
            raise ValueError("`xuids` cannot be empty")
        url = f"{_HOST}/hi/playlist/{playlist_id}/csrs"
        params: dict = {"players": [wrap_xuid(x) for x in xuids]}
        if season_id:
            params["season"] = _clean_season_id(season_id)
        return PlaylistCsr(**await self._get_json(url, params=params))


def _clean_season_id(season_id: str) -> str:
    """Remove the path prefix and ".json" extension from a season ID.

    For example, season "Csr/Seasons/CsrSeason5-1.json" is cleaned to
    "CsrSeason5-1".
    """
    return Path(season_id).stem



"""Test the spnkr.models.types module."""

import pytest
from pydantic import BaseModel

from spnkr.models import types


def test_read_only_dict_getitem():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert rod["foo"] == "bar"


def test_read_only_dict_len():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert len(rod) == 1


def test_read_only_dict_iter():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert list(rod) == ["foo"]


def test_read_only_dict_str():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert str(rod) == "{'foo': 'bar'}"


def test_read_only_dict_repr():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert repr(rod) == "{'foo': 'bar'}"


def test_read_only_dict_setitem():
    rod = types._ReadOnlyDict({"foo": "bar"})
    with pytest.raises(TypeError):
        rod["foo"] = "baz"  # type: ignore


def test_read_only_dict_delitem():
    rod = types._ReadOnlyDict({"foo": "bar"})
    with pytest.raises(TypeError):
        del rod["foo"]  # type: ignore


def test_read_only_dict_to_dict():
    rod = types._ReadOnlyDict({"foo": "bar"})
    assert dict(rod) == {"foo": "bar"}


def test_read_only_dict_annotation():
    class Foo(BaseModel):
        bar: types.ReadOnlyDict[str, str]

    foo = Foo(bar={"foo": "bar"})
    assert isinstance(foo.bar, types._ReadOnlyDict)
    assert foo.bar["foo"] == "bar"
    assert foo.model_dump() == {"bar": {"foo": "bar"}}
    assert foo.model_dump_json() == '{"bar":{"foo":"bar"}}'



"""Test SkillService."""

import pytest

from spnkr.services.skill import SkillService, _clean_season_id


@pytest.fixture
def service(session):
    return SkillService(session)


@pytest.mark.asyncio
async def test_get_match_skill(session, service: SkillService):
    session.set_response("get_match_skill.json")
    await service.get_match_skill("match_id", [1234567890123456])
    session.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/matches/match_id/skill",
        params={"players": ["xuid(1234567890123456)"]},
    )


@pytest.mark.asyncio
async def test_get_match_skill_str(service: SkillService):
    with pytest.raises(TypeError):
        await service.get_match_skill("match_id", "1234567890123456")


@pytest.mark.asyncio
async def test_get_match_skill_empty(service: SkillService):
    with pytest.raises(ValueError):
        await service.get_match_skill("match_id", [])


@pytest.mark.asyncio
async def test_get_playlist_csr(session, service: SkillService):
    session.set_response("get_playlist_csr.json")
    await service.get_playlist_csr(
        "playlist_id", [1234567890123456, 2345678901234567], "CsrSeason5-1"
    )
    session.get.assert_called_with(
        "https://skill.svc.halowaypoint.com:443/hi/playlist/playlist_id/csrs",
        params={
            "players": ["xuid(1234567890123456)", "xuid(2345678901234567)"],
            "season": "CsrSeason5-1",
        },
    )


@pytest.mark.asyncio
async def test_get_playlist_csr_str(service: SkillService):
    with pytest.raises(TypeError):
        await service.get_playlist_csr("match_id", "1234567890123456")


@pytest.mark.asyncio
async def test_get_playlist_csr_empty(service: SkillService):
    with pytest.raises(ValueError):
        await service.get_playlist_csr("match_id", [])


def test_clean_season_id():
    assert _clean_season_id("Csr/Seasons/CsrSeason5-1.json") == "CsrSeason5-1"
    assert _clean_season_id(" csr/seasons/csrseason5-1.json ") == "csrseason5-1"
    assert _clean_season_id("CsrseasoN5-1") == "CsrseasoN5-1"



"""Test GameContentHacsService."""

import pytest

from spnkr.services.gamecms_hacs import GameCmsHacsService


@pytest.fixture
def service(session):
    return GameCmsHacsService(session)


@pytest.mark.asyncio
async def test_get_medal_metadata(session, service: GameCmsHacsService):
    session.set_response("get_medal_metadata.json")
    await service.get_medal_metadata()
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/Waypoint/file/medals/metadata.json"
    )


@pytest.mark.asyncio
async def test_get_csr_season_calendar(session, service: GameCmsHacsService):
    session.set_response("get_csr_season_calendar.json")
    await service.get_csr_season_calendar()
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/Progression/file/Csr/Calendars/CsrSeasonCalendar.json"
    )


@pytest.mark.asyncio
async def test_get_season_calendar(session, service: GameCmsHacsService):
    session.set_response("get_season_calendar.json")
    await service.get_season_calendar()
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/progression/file/calendars/seasons/seasoncalendar.json"
    )


@pytest.mark.asyncio
async def test_get_career_reward_track(session, service: GameCmsHacsService):
    session.set_response("get_career_reward_track.json")
    await service.get_career_reward_track()
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/Progression/file/RewardTracks/CareerRanks/careerRank1.json"
    )


@pytest.mark.asyncio
async def test_get_image(session, service: GameCmsHacsService):
    session.set_response("career_rank_corporal_gold_III.png")
    result = await service.get_image("path/to/image.png")
    assert round(len(result) / 1024) == 15
    session.get.assert_called_with(
        "https://gamecms-hacs.svc.halowaypoint.com/hi/images/file/path/to/image.png"
    )



"""Test DiscoveryUgcService."""

import datetime as dt

import pytest

from spnkr.services.discovery_ugc import DiscoveryUgcService


@pytest.fixture
def service(session):
    return DiscoveryUgcService(session)


@pytest.mark.asyncio
async def test_get_map_mode_pair(session, service: DiscoveryUgcService):
    session.set_response("get_map_mode_pair.json")
    await service.get_map_mode_pair("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/mapModePairs/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_playlist(session, service: DiscoveryUgcService):
    session.set_response("get_playlist.json")
    await service.get_playlist("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/playlists/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_map(session, service: DiscoveryUgcService):
    session.set_response("get_map.json")
    await service.get_map("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/maps/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_get_ugc_game_variant(session, service: DiscoveryUgcService):
    session.set_response("get_ugc_game_variant.json")
    await service.get_ugc_game_variant("asset_id", "version_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/ugcGameVariants/asset_id/versions/version_id"
    )


@pytest.mark.asyncio
async def test_search_assets_default(session, service: DiscoveryUgcService):
    session.set_response("search_assets.json")
    await service.search_assets(
        start=0, count=10, sort="plays_recent", order="desc"
    )
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/search",
        params={
            "start": 0,
            "count": 10,
            "sort": "playsrecent",
            "order": "desc",
        },
    )


@pytest.mark.asyncio
async def test_search_assets_all(session, service: DiscoveryUgcService):
    session.set_response("search_assets.json")
    await service.search_assets(
        start=0,
        count=10,
        sort="plays_recent",
        order="desc",
        asset_kind="map",
        term="Live Fire",
        tags=["343i"],
        author="aaid(123)",
        average_rating_min=4.0,
        from_date_created_utc=dt.date(2021, 1, 1),
        to_date_created_utc=dt.date(2021, 1, 31),
        from_date_modified_utc=dt.date(2021, 1, 1),
        to_date_modified_utc=dt.date(2021, 1, 31),
        from_date_published_utc=dt.date(2021, 1, 1),
        to_date_published_utc=dt.date(2021, 1, 31),
    )
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/search",
        params={
            "start": 0,
            "count": 10,
            "sort": "playsrecent",
            "order": "desc",
            "assetKind": "map",
            "term": "Live Fire",
            "tags": ["343i"],
            "author": "aaid(123)",
            "averageRatingMin": 4.0,
            "fromDateCreatedUtc": "2021-01-01",
            "toDateCreatedUtc": "2021-01-31",
            "fromDateModifiedUtc": "2021-01-01",
            "toDateModifiedUtc": "2021-01-31",
            "fromDatePublishedUtc": "2021-01-01",
            "toDatePublishedUtc": "2021-01-31",
        },
    )


@pytest.mark.asyncio
async def test_search_assets_invalid_count(service: DiscoveryUgcService):
    with pytest.raises(ValueError):
        await service.search_assets(start=0, count=0)
    with pytest.raises(ValueError):
        await service.search_assets(start=0, count=102)


@pytest.mark.asyncio
async def test_get_film_by_match_id(session, service: DiscoveryUgcService):
    session.set_response("get_film_by_match_id.json")
    await service.get_film_by_match_id("match_id")
    session.get.assert_called_with(
        "https://discovery-infiniteugc.svc.halowaypoint.com:443/hi/films/matches/match_id/spectate"
    )



import json
from pathlib import Path
from typing import Any
from uuid import UUID

import pytest

from spnkr.models.discovery_ugc import (
    AssetSearchPage,
    Film,
    Map,
    MapModePair,
    Playlist,
    UgcGameVariant,
)
from spnkr.models.gamecms_hacs import (
    CareerRewardTrack,
    CsrSeasonCalendar,
    MedalMetadata,
    SeasonCalendar,
)
from spnkr.models.profile import User
from spnkr.models.skill import MatchSkill, PlaylistCsr
from spnkr.models.stats import (
    MatchCount,
    MatchHistory,
    MatchStats,
    ServiceRecord,
)

RESPONSES = Path("tests/responses")


def load_response(name: str) -> Any:
    with open(RESPONSES / f"{name}.json", encoding="utf-8") as f:
        return json.load(f)


def test_parse_medal_metadata():
    data = load_response("get_medal_metadata")
    result = MedalMetadata(**data)
    assert result.medals[0].name_id == 622331684


def test_parse_csr_season_calendar():
    data = load_response("get_csr_season_calendar")
    result = CsrSeasonCalendar(**data)
    assert len(result.seasons) == 8


def test_parse_season_calendar():
    data = load_response("get_season_calendar")
    result = SeasonCalendar(**data)
    assert len(result.events) == 30


def test_parse_career_reward_track():
    data = load_response("get_career_reward_track")
    result = CareerRewardTrack(**data)
    assert len(result.ranks) == 272
    assert result.name.value == "Career Rank"


def test_parse_match_skill():
    data = load_response("get_match_skill")
    result = MatchSkill(**data)
    assert result.value[0].id == "xuid(2535445291321133)"


def test_parse_playlist_csr():
    data = load_response("get_playlist_csr")
    result = PlaylistCsr(**data)
    assert result.value[0].id == "xuid(2535445291321133)"


def test_parse_match_count():
    data = load_response("get_match_count")
    result = MatchCount(**data)
    assert result.matchmade_matches_played_count == 729


def test_parse_service_record():
    data = load_response("get_service_record")
    result = ServiceRecord(**data)
    assert result.matches_completed == 1915


def test_parse_service_record_empty():
    data = load_response("get_service_record_empty")
    result = ServiceRecord(**data)
    assert result.matches_completed == 0


def test_parse_match_history():
    data = load_response("get_match_history")
    result = MatchHistory(**data)
    expected = UUID("a1219f5c-5942-4cd5-9ff1-4b6ea89905bc")
    assert result.results[0].match_id == expected


def test_parse_match_stats():
    data = load_response("get_match_stats")
    result = MatchStats(**data)
    expected = UUID("6f050134-bede-47bc-a6df-eeafdcb9f97f")
    assert result.match_id == expected


def test_parse_match_stats_ctf():
    data = load_response("get_match_stats_ctf")
    result = MatchStats(**data)
    expected = UUID("ffccbba6-fd52-4ff8-aa41-4bdc6487cd92")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.capture_the_flag_stats
    assert mode is not None


def test_parse_match_stats_zones():
    data = load_response("get_match_stats_zones")
    result = MatchStats(**data)
    expected = UUID("ff9af871-ea6b-40c4-abe6-9aab6bfe1808")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.zones_stats
    assert mode is not None


def test_parse_match_stats_elimination():
    data = load_response("get_match_stats_elimination")
    result = MatchStats(**data)
    expected = UUID("fe368e0b-9281-43ad-9b3d-9b16fe1e9402")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.elimination_stats
    assert mode is not None


def test_parse_match_stats_oddball():
    data = load_response("get_match_stats_oddball")
    result = MatchStats(**data)
    expected = UUID("fe75a257-acd6-41ac-a778-8f86765639cb")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.oddball_stats
    assert mode is not None


def test_parse_match_stats_stockpile():
    data = load_response("get_match_stats_stockpile")
    result = MatchStats(**data)
    expected = UUID("f40f4aa4-cc28-466d-b4ec-86fd86424083")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.stockpile_stats
    assert mode is not None


def test_parse_match_stats_extraction():
    data = load_response("get_match_stats_extraction")
    result = MatchStats(**data)
    expected = UUID("44b9ca37-1d91-438a-8ef1-aa4cb78a19b2")
    assert result.match_id == expected
    mode = result.players[0].player_team_stats[0].stats.extraction_stats
    assert mode is not None


def test_parse_game_variant():
    data = load_response("get_ugc_game_variant")
    result = UgcGameVariant(**data)
    assert result.public_name == "Fiesta:Slayer"


def test_parse_map_mode_pair():
    data = load_response("get_map_mode_pair")
    result = MapModePair(**data)
    assert result.public_name == "Arena:Shotty Snipes Slayer on Catalyst"


def test_parse_map():
    data = load_response("get_map")
    result = Map(**data)
    assert result.public_name == "Forge Space"


def test_parse_playlist():
    data = load_response("get_playlist")
    result = Playlist(**data)
    assert result.public_name == "Ranked Arena"


def test_parse_asset_search_page():
    data = load_response("search_assets")
    result = AssetSearchPage(**data)
    assert len(result.results) == 10


def test_parse_film():
    data = load_response("get_film_by_match_id")
    result = Film(**data)
    assert len(result.custom_data.chunks) == 27


def test_film_get_chunks_and_urls():
    data = load_response("get_film_by_match_id")
    result = Film(**data)
    chunk_urls = result.get_chunks_and_urls()
    assert len(chunk_urls) == 27
    assert chunk_urls[0][0].chunk_type == 1  # Chunk header
    assert chunk_urls[0][1] == (
        "https://blobs-infiniteugc.svc.halowaypoint.com/ugcstorage/film/3a86cc7e-4a15-49a0-b17d-1b1cdf16c49f/86ab777c-895d-4237-9fa0-43c7932322ac/filmChunk0"
    )


def test_parse_user():
    data = load_response("get_user")
    result = User(**data)
    assert result.xuid == 1234567890


def test_parse_users():
    data = load_response("get_users")
    result = [User(**user) for user in data]
    assert len(result) == 1
    assert result[0].xuid == 1234567890


if __name__ == "__main__":
    pytest.main()
